from .render import RenderWalker


def dumps(tree, strict=False):
    return Dumper(strict=strict).dump(tree)


class Dumper(RenderWalker):
    def before_string(self, string, key):
        self.dump += string

    def before_constant(self, constant, key):
        self.dump += constant

    def dump(self, tree):
        self.dump = ''
        self.walk(tree)
        return self.dump



import sys
import re


python_version = sys.version_info[0]
python_subversion = sys.version_info[1]
string_instance = str if python_version == 3 else basestring


class BaronError(Exception):
    pass


class FlexibleIterator():
    def __init__(self, sequence):
        self.sequence = sequence
        self.position = -1

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        self.position += 1
        if self.position == len(self.sequence):
            raise StopIteration
        return self.sequence[self.position]

    def next_starts_with(self, sentence):
        size_of_choice = len(sentence)
        return self.sequence[self.position + 1: self.position + 1 + size_of_choice] == sentence

    def next_in(self, choice):
        if self.position + 1 >= len(self.sequence):
            return False
        return self.sequence[self.position + 1] in choice

    def show_next(self, at=1):
        if self.position + at >= len(self.sequence):
            return None
        return self.sequence[self.position + at]

    def rest_of_the_sequence(self):
        return self.sequence[self.position + 1:]

    def end(self):
        return self.position == (len(self.sequence) - 1)

    def grab(self, test):
        to_return = ""
        current = None
        while self.show_next() is not None and test(self):
            current = next(self)
            to_return += current

        return to_return

    def grab_string(self, test):
        to_return = ""
        current = None
        escaped = False
        while self.show_next() is not None and (escaped or test(self)):
            current = next(self)
            to_return += current
            if escaped:
                escaped = False
            elif current == "\\":
                escaped = True

        return to_return


def create_node_from_token(token, **kwargs):
    result = {"type": token.name.lower(), "value": token.value}
    if kwargs:
        result.update(kwargs)
    return result


def create_node(name, value, **kwargs):
    result = {"type": name, "value": value}
    if kwargs:
        result.update(kwargs)
    return result


newline_regex = re.compile("(\r\n|\n|\r)")


def is_newline(text):
    return newline_regex.match(text)


def split_on_newlines(text):
    newlines = newline_regex.finditer(text)
    if not newlines:
        yield text
    else:
        current_position = 0
        for newline in newlines:
            yield text[current_position:newline.start(1)]
            yield text[newline.start(1):newline.end(1)]
            current_position = newline.end(1)
        yield text[current_position:]


# Thanks to
# https://github.com/nvie/rq/commit/282f4be9316d608ebbacd6114aab1203591e8f95
if python_version >= 3 or python_subversion >= 7:
    from functools import total_ordering
else:
    def total_ordering(cls):
        """Class decorator that fills in missing ordering methods"""
        convert = {
            '__lt__': [('__gt__', lambda self, other: other < self),
                       ('__le__', lambda self, other: not other < self),
                       ('__ge__', lambda self, other: not self < other)],
            '__le__': [('__ge__', lambda self, other: other <= self),
                       ('__lt__', lambda self, other: not other <= self),
                       ('__gt__', lambda self, other: not self <= other)],
            '__gt__': [('__lt__', lambda self, other: other > self),
                       ('__ge__', lambda self, other: not other > self),
                       ('__le__', lambda self, other: not self > other)],
            '__ge__': [('__le__', lambda self, other: other >= self),
                       ('__gt__', lambda self, other: not other >= self),
                       ('__lt__', lambda self, other: not self >= other)]
        }
        roots = set(dir(cls)) & set(convert)
        if not roots:
            raise ValueError('must define at least one ordering operation: < > <= >=')  # noqa
        root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
        for opname, opfunc in convert[root]:
            if opname not in roots:
                opfunc.__name__ = opname
                opfunc.__doc__ = getattr(int, opname).__doc__
                setattr(cls, opname, opfunc)
        return cls



#!/usr/bin/python
# -*- coding:Utf-8 -*-

import baron
from .test_utils import check_dumps


def test_empty():
    check_dumps("")


def test_var():
    check_dumps("hello")


def test_int():
    check_dumps("1")


def test_assign():
    check_dumps("a = 2")


def test_assign_star_expr():
    check_dumps("a, *b = 2")
    check_dumps("a, *  b = 2")


def test_standalone_annotation():
    check_dumps("a :  str")


def test_assign_annotation():
    check_dumps("a :  int   =      2")


def test_binary_operator():
    check_dumps("z +  42")
    check_dumps("z   -  42")


def test_while():
    check_dumps("while a  : pass\n")


def test_while_else():
    check_dumps("while a  : pass\nelse : pass\n")


def test_while_indent():
    check_dumps("while a:\n    pass\n")


def test_if():
    check_dumps("if a:\n    pass\n")


def test_if_elif():
    check_dumps("if a: \n    pass\nelif b: pass\n")


def test_if_elif_else():
    check_dumps("if a: \n    pass\nelif b: pass\nelse :   \n	pouet\n")


def test_import():
    check_dumps("import  a")


def test_import_madness():
    check_dumps("import  a.B .   c as  saucisse")


def test_from_import():
    check_dumps("from b   import  a  as   rev")


def test_from_import_special_notation():
    check_dumps("from a import (b)")
    check_dumps("from a import (b, c, d)")


def test_print_empty():
    check_dumps("print")


def test_print():
    check_dumps("print pouet")


def test_print_madness():
    check_dumps("print >>  qsd, pouet, zdzd,")


def test_atom_trailers_call():
    check_dumps("a.c(b)")


def test_atom_trailers_call_default():
    check_dumps("caramba(s, b=2)")


def test_list_argument():
    check_dumps("caramba(* a)")


def test_dict_argument():
    check_dumps("caramba(** a)")


def test_list_argument_funcdef():
    check_dumps("def caramba(* a): pass\n")


def test_dict_argument_funcdef():
    check_dumps("def caramba(** a): pass\n")


def test_funcdef_return_annotation():
    check_dumps("def caramba() -> Stuff: pass\n")


def test_string():
    check_dumps("'ama string!'")


def test_funcdef():
    check_dumps("def a  ( ) : pass\n")


def test_funcdef_indent():
    check_dumps("def a  ( ) : \n    pass\n")


def test_funcdef_parameter():
    check_dumps("def a  ( b ) : pass\n")


def test_funcdef_parameter_named():
    check_dumps("def a  ( b  , c = qsd ) : pass\n")


def test_funcdef_parameter_named_star():
    check_dumps("def a  ( b  , c = qsd , * d ) : pass\n")


def test_funcdef_parameter_named_star_start():
    check_dumps("def a  ( b  , c = qsd , * d , ** e ) : pass\n")


def test_funcdef_typed_parameter():
    check_dumps("def a  ( b : c ) : pass\n")


def test_funcdef_typed_parameter_named():
    check_dumps("def a  ( b  , c : d = qsd ) : pass\n")


def test_funcdef_typed_parameter_named_star():
    check_dumps("def a  ( b  , c = qsd , * d : qsd ) : pass\n")


def test_funcdef_typed_parameter_named_star_start():
    check_dumps("def a  ( b  , c = qsd , * d , ** e : qsd ) : pass\n")


def test_async_funcdef():
    check_dumps("async def a  ( ) : pass\n")


def test_async_funcdef_indent():
    check_dumps("async def a  ( ) : \n    pass\n")


def test_async_funcdef_parameter():
    check_dumps("async def a  ( b ) : pass\n")


def test_async_funcdef_parameter_named():
    check_dumps("async def a  ( b  , c = qsd ) : pass\n")


def test_return():
    check_dumps("return a")


def test_getitem():
    check_dumps("a[ b  ]")


def test_slice_empty():
    check_dumps("a[ :  ]")


def test_slice_classical():
    check_dumps("a[1: 42]")


def test_slice_step():
    check_dumps("a[1: 42:]")
    check_dumps("a[1: 42    :         3]")


def test_unitary_operator():
    check_dumps("- 1")


def test_unicode_string():
    check_dumps("u'pouet'")


def test_raw_string():
    check_dumps("r'pouet'")


def test_unicode_raw_string():
    check_dumps("ur'pouet'")


def test_binary_string():
    check_dumps("b'pouet'")


def test_binary_raw_string():
    check_dumps("br'pouet'")


def test_interpolated_string():
    check_dumps("f'qsd'")


def test_nterporlated_raw_string():
    check_dumps("rf'qsd'")


def test_for():
    check_dumps("for i in pouet : pass\n")


def test_for_indent():
    check_dumps("for i in pouet : \n    pass\n")


def test_for_else():
    check_dumps("for i in pouet : pass\nelse: pass\n")


def test_async_for():
    check_dumps("async for i in pouet : pass\n")


def test_async_for_indent():
    check_dumps("async for i in pouet : \n    pass\n")


def test_async_for_else():
    check_dumps("async for i in pouet : pass\nelse: pass\n")


def test_lambda():
    check_dumps("lambda : x")


def test_lambda_args():
    check_dumps("lambda poeut, hompi_dompi: x")


def test_try_finally():
    check_dumps("try : pass\nfinally : pass\n")


def test_try_except():
    check_dumps("try : pass\nexcept Exception : pass\n")


def test_try_except_comma():
    check_dumps("try : pass\nexcept Exception ,   d : pass\n")


def test_try_except_as():
    check_dumps("try : pass\nexcept Exception     as   d : pass\n")


def test_try_except_finally():
    check_dumps("try : pass\nexcept Exception : pass\nfinally : pass\n")


def test_try_except_finally_else():
    check_dumps("try : pass\nexcept Exception : pass\nelse: pouet\nfinally : pass\n")


def test_comment():
    check_dumps("# pouet")


def test_comment_formatting():
    check_dumps("a # pouet")


def test_augassign():
    check_dumps("a &= b")


def test_call_fourth_formatting():
    check_dumps("set(n for e in s() \n                 if a)")


def test_boolean_operator():
    check_dumps("a and b")


def test_boolean_operator_advanced():
    check_dumps("a and b or c and d")


def test_comparison():
    check_dumps("a < b")


def test_with():
    check_dumps("with a : \n    pass\n")


def test_with_as():
    check_dumps("with a as b : \n    pass\n")


def test_async_with():
    check_dumps("async with a : \n    pass\n")


def test_async_with_as():
    check_dumps("async with a as b : \n    pass\n")


def test_dict_empty():
    check_dumps("{   }")


def test_dict_one():
    check_dumps("{ a : b  }")


def test_dict_more():
    check_dumps("{ a : b   ,\n123  :     'pouet'  }")


def test_ternary_operator():
    check_dumps("a   if        b  else      c")


def test_yield_empty():
    check_dumps("yield")


def test_yield():
    check_dumps("yield a")


def test_decorator():
    check_dumps("@pouet\ndef a(): pass\n")


def test_decorator_call():
    check_dumps("@pouet('pouet')\ndef a(): pass\n")


def test_class():
    check_dumps("class A: pass\n")


def test_class_parenthesis():
    check_dumps("class A(): pass\n")


def test_class_parenthesis_inherit():
    check_dumps("class A(B): pass\n")


def test_class_parenthesis_inherit_arglist():
    check_dumps("class A(**B): pass\n")


def test_class_parenthesis_inherit_decorated():
    check_dumps("@pouet\nclass A(B): pass\n")


def test_tuple():
    check_dumps("a  ,  b    , c")


def test_tuple_parenthesis():
    check_dumps("( a  ,  b    , c    )")


def test_return_empty():
    check_dumps("return")


def test_list_empty():
    check_dumps("[   ]")


def test_list():
    check_dumps("[ x ]")


def test_list_more():
    check_dumps("[ x, r, f, e   , e ]")


def test_associative_parenthesis():
    check_dumps("( \n   ( a ) +   ( 1 *    4 )\n ) ")


def test_fplist():
    check_dumps("def a((b, c)): pass\n")


def test_break():
    check_dumps("break")


def test_assert():
    check_dumps("assert a == b")
    check_dumps("assert a == b  , c")


def test_continue():
    check_dumps("continue")


def test_raise():
    check_dumps("raise")
    check_dumps("raise a")
    check_dumps("raise a ,   b")
    check_dumps("raise a   from      b")
    check_dumps("raise a ,   b   ,     c")


def test_del():
    check_dumps("del a")


def test_star():
    check_dumps("from a import *")


def test_string_chain():
    check_dumps("'q' 'b'")


def test_list_comprehension():
    check_dumps("[ x for   y       in  z      ]")


def test_list_comprehension_ifs():
    check_dumps("[ x for   y       in  z   if a   if  qsd  ]")


def test_list_comprehension_ifs_more():
    check_dumps("[ x for   y       in  z   if a   if  qsd  for ss in gfgr    ]")


def test_generator_comprehension():
    check_dumps("( x for   y       in  z      )")


def test_generator_comprehension_ifs():
    check_dumps("( x for   y       in  z   if a   if  qsd  )")


def test_generator_comprehension_ifs_more():
    check_dumps("( x for   y       in  z   if a   if  qsd  for ss in gfgr    )")


def test_dict_comprehension():
    check_dumps("{ x: z for   y       in  z      }")


def test_dict_comprehension_ifs():
    check_dumps("{ x   : z for   y       in  z   if a   if  qsd  }")


def test_dict_comprehension_ifs_more():
    check_dumps("{ x :z for   y       in  z   if a   if  qsd  for ss in gfgr    }")


def test_set_comprehension():
    check_dumps("{ x for   y       in  z      }")


def test_set_comprehension_ifs():
    check_dumps("{ x    for   y       in  z   if a   if  qsd  }")


def test_set_comprehension_ifs_more():
    check_dumps("{ x  for   y       in  z   if a   if  qsd  for ss in gfgr    }")


def test_set_one():
    check_dumps("{a}")


def test_set_one_comma():
    check_dumps("{a,}")


def test_set_one_more():
    check_dumps("{ a  ,   b      }")


def test_argument_generator_comprehension():
    check_dumps("a( x for   y       in  z)")


def test_argument_generator_comprehension_comprehension_ifs():
    check_dumps("a( x    for   y       in  z   if a   if  qsd)")


def test_argument_generator_comprehension_comprehension_ifs_more():
    check_dumps("a(x  for   y       in  z   if a   if  qsd  for ss in gfgr)")


def test_hexa():
    check_dumps("0x123")


def test_octa():
    check_dumps("0123")


def test_binary():
    check_dumps("0b010110101100")


def test_float():
    check_dumps("1.2")


def test_complex():
    check_dumps("10j")


def test_float_exponant():
    check_dumps("1e9")


def test_semicolon():
    check_dumps("a;b")


def test_exec():
    check_dumps("exec a")


def test_exec_globals():
    check_dumps("exec a  in   b")


def test_exec_globals_locals():
    check_dumps("exec a  in   b   ,     c")


def test_exec_function():
    check_dumps("exec(a)")


def test_global():
    check_dumps("global a")


def test_global_more():
    check_dumps("global a ,   b,    d")


def test_ellipsis():
    check_dumps("a[ ... ]")


def test_yield_atom_empty():
    check_dumps("a = ( yield )")


def test_yield_atom():
    check_dumps("a = ( yield b )")


def test_repr():
    check_dumps("` a  `")


def test_comment_special_case():
    check_dumps("d((s)   # Padding\n)")


def test_from_import_parenthesis_formatting():
    check_dumps("from a import (\nb, c\n)\n")


def test_getitem_special_case():
    check_dumps("[a[...] ]")


def test_print_tuple():
    check_dumps("print(pouet, pouet)")


def test_raise_special():
    check_dumps("raise   # pouet")


def test_from_import_star_comment():
    check_dumps("from a import * # pouet")


def test_class_formatting():
    check_dumps("class A(pouet) : \n    pass\n")


def test_backslash_not_in_formatting():
    check_dumps("if a not \\\n      in b: pass\n")


def test_try_import_after_colon():
    check_dumps("try: import stuff\nexcept: pass\n")


def test_single_object():
    assert baron.dumps({"type": "name", "value": "a"}) == "a"


def test_crash_issue_85():
    check_dumps('d*e-1\n')


def test_keyword_only_marker():
    check_dumps("def foo(a, *, b, c):    pass\n")
    check_dumps("def foo(a, *  , b, c):    pass\n")


def test_float_with_underscores():
    check_dumps("123_456.789_012")



#!/usr/bin/python
# -*- coding:Utf-8 -*-


from baron.tokenizer import tokenize, KEYWORDS


def match(string, token):
    assert tokenize([string]) == [(token, string), ('ENDMARKER', ''), None]


def test_empty():
    assert tokenize([]) == [('ENDMARKER', ''), None]


def test_name():
    match('a', 'NAME')
    match('pouet', 'NAME')


def test_name__():
    match('_a', 'NAME')


def test_name_number():
    match('a123', 'NAME')


def test_await_as_name():
    # not a real keyword in python
    match('await', 'NAME')


def test_asyc_as_name():
    # not a real keyword in python
    match('async', 'NAME')


def test_number():
    match('1234', 'INT')
    match('1234L', 'LONG')
    match('1234l', 'LONG')


def test_float():
    match('1234.', 'FLOAT')
    match('.1234', 'FLOAT')
    match('1234.1234', 'FLOAT')
    match('1234.L', 'FLOAT')
    match('.1234L', 'FLOAT')
    match('1234.1234L', 'FLOAT')
    match('1234.l', 'FLOAT')
    match('.1234l', 'FLOAT')
    match('1234.1234l', 'FLOAT')


def test_hexa():
    match('0x123ABCD', 'HEXA')
    match('0X123ABCD', 'HEXA')
    match('0x123ABCDl', 'HEXA')
    match('0x123ABCDL', 'HEXA')


def test_grouped():
    match('100_000', 'INT')
    match('10_000_000.', 'FLOAT')
    match('10_000_000.0', 'FLOAT')
    match('0xCAFE_F00D', 'HEXA')
    match('0o123_45', 'OCTA')
    match('0b_0011_1111_0100_1110', 'BINARY')


def test_octa():
    match('012345', 'OCTA')
    match('0o12345', 'OCTA')
    match('0O1235', 'OCTA')
    match('0O1235l', 'OCTA')
    match('0O1235L', 'OCTA')


def test_float_exponant():
    match('1234e10', 'FLOAT_EXPONANT')
    match('1234E10', 'FLOAT_EXPONANT')
    match('1234e+10', 'FLOAT_EXPONANT')
    match('1234e-10', 'FLOAT_EXPONANT')
    match('1234E+10', 'FLOAT_EXPONANT')
    match('1234E-10', 'FLOAT_EXPONANT')
    match('1234.5678e10', 'FLOAT_EXPONANT')
    match('1234.5678E10', 'FLOAT_EXPONANT')
    match('1234.5678e+10', 'FLOAT_EXPONANT')
    match('1234.5678e-10', 'FLOAT_EXPONANT')
    match('1234.5678E+10', 'FLOAT_EXPONANT')
    match('1234.5678E-10', 'FLOAT_EXPONANT')
    match('.5678e10', 'FLOAT_EXPONANT')
    match('.5678E10', 'FLOAT_EXPONANT')
    match('.5678e+10', 'FLOAT_EXPONANT')
    match('.5678e-10', 'FLOAT_EXPONANT')
    match('.5678E+10', 'FLOAT_EXPONANT')
    match('.5678E-10', 'FLOAT_EXPONANT')


def test_floating_point_parser_bug_85():
    from baron import parse
    assert parse('d*e-1') == [
        {'first': {'first': {'type': 'name', 'value': 'd'},
                   'first_formatting': [],
                   'second': {'type': 'name', 'value': 'e'},
                   'second_formatting': [],
                   'type': 'binary_operator',
                   'value': '*'},
         'first_formatting': [],
         'second': {'section': 'number', 'type': 'int', 'value': '1'},
         'second_formatting': [],
         'type': 'binary_operator',
         'value': '-'}]


def test_left_parenthesis():
    match('(', 'LEFT_PARENTHESIS')


def test_right_parenthesis():
    match(')', 'RIGHT_PARENTHESIS')


def test_colon():
    match(':', 'COLON')
    assert tokenize([':']) == [('COLON', ':'), ('ENDMARKER', ''), None]


def test_comma():
    match(',', 'COMMA')


def test_semicolon():
    match(';', 'SEMICOLON')


def test_sequence():
    assert tokenize(['a', '123']) == [('NAME', 'a'), ('INT', '123'), ('ENDMARKER', ''), None]


def test_plus():
    match('+', 'PLUS')


def test_minus():
    match('-', 'MINUS')


def test_start():
    match('*', 'STAR')


def test_slash():
    match('/', 'SLASH')


def test_vbar():
    match('|', 'VBAR')


def test_amper():
    match('&', 'AMPER')


def test_less():
    match('<', 'LESS')


def test_greater():
    match('>', 'GREATER')


def test_equal():
    match('=', 'EQUAL')


def test_dot():
    match('.', 'DOT')


def test_percent():
    match('%', 'PERCENT')


def test_left_square_bracket():
    match('[', 'LEFT_SQUARE_BRACKET')


def test_right_square_bracket():
    match(']', 'RIGHT_SQUARE_BRACKET')


def test_left_bracket():
    match('{', 'LEFT_BRACKET')


def test_right_bracket():
    match('}', 'RIGHT_BRACKET')


def test_back_quote():
    match('`', 'BACKQUOTE')


def test_equal_equal():
    match('==', 'EQUAL_EQUAL')


def test_not_equal():
    match('!=', 'NOT_EQUAL')
    match('<>', 'NOT_EQUAL')


def test_less_equal():
    match('<=', 'LESS_EQUAL')


def test_greater_equal():
    match('>=', 'GREATER_EQUAL')


def test_tilde():
    match('~', 'TILDE')


def test_circumflex():
    match('^', 'CIRCUMFLEX')


def test_left_shift():
    match('<<', 'LEFT_SHIFT')


def test_right_shift():
    match('>>', 'RIGHT_SHIFT')


def test_double_star():
    match('**', 'DOUBLE_STAR')


def test_plus_equal():
    match('+=', 'PLUS_EQUAL')


def test_minus_equal():
    match('-=', 'MINUS_EQUAL')


def test_star_equal():
    match('*=', 'STAR_EQUAL')


def test_slash_equal():
    match('/=', 'SLASH_EQUAL')


def test_percent_equal():
    match('%=', 'PERCENT_EQUAL')


def test_amper_equal():
    match('&=', 'AMPER_EQUAL')


def test_at_equal():
    match('@=', 'AT_EQUAL')


def test_vbar_equal():
    match('|=', 'VBAR_EQUAL')


def test_circumflex_equal():
    match('^=', 'CIRCUMFLEX_EQUAL')


def test_left_shift_equal():
    match('<<=', 'LEFT_SHIFT_EQUAL')


def test_right_shift_equal():
    match('>>=', 'RIGHT_SHIFT_EQUAL')


def test_double_star_equal():
    match('**=', 'DOUBLE_STAR_EQUAL')


def test_double_slash():
    match('//', 'DOUBLE_SLASH')


def test_double_slash_equal():
    match('//=', 'DOUBLE_SLASH_EQUAL')


def test_keywords():
    for keyword in KEYWORDS:
        assert tokenize([keyword]) == [(keyword.upper(), keyword), ('ENDMARKER', ''), None]


def test_endl():
    match('\n', 'ENDL')
    match('\r\n', 'ENDL')


def test_simple_string():
    match('"pouet pouet"', 'STRING')
    match("'pouet pouet'", "STRING")


def test_multi_string():
    match('"""pouet pouet"""', 'STRING')
    match("'''pouet pouet'''", "STRING")


def test_unicode_string():
    match('u"pouet pouet"', 'UNICODE_STRING')
    match("u'pouet pouet'", "UNICODE_STRING")
    match('u"""pouet pouet"""', 'UNICODE_STRING')
    match("u'''pouet pouet'''", "UNICODE_STRING")
    match('U"pouet pouet"', 'UNICODE_STRING')
    match("U'pouet pouet'", "UNICODE_STRING")
    match('U"""pouet pouet"""', 'UNICODE_STRING')
    match("U'''pouet pouet'''", "UNICODE_STRING")


def test_interpolated_string():
    match("f'He said his name is {name!r}.'", 'INTERPOLATED_STRING')
    match("f'The value is {value}.'", 'INTERPOLATED_STRING')
    match('F"He said his name is {name!r}."', 'INTERPOLATED_STRING')
    match('f"The value is {value}."', 'INTERPOLATED_STRING')
    match("F'{date} was on a {date:%A}'", 'INTERPOLATED_STRING')
    match("f'a={d[\"a\"]}'", 'INTERPOLATED_STRING')
    match("F'{(lambda x: x*2)(3)}'", 'INTERPOLATED_STRING')


def test_interpolated_raw_string():
    match("fr'He said his name is {name!r}.'", 'INTERPOLATED_RAW_STRING')
    match("fr'The value is {value}.'", 'INTERPOLATED_RAW_STRING')
    match('Fr"He said his name is {name!r}."', 'INTERPOLATED_RAW_STRING')
    match('fR"The value is {value}."', 'INTERPOLATED_RAW_STRING')
    match("FR'{date} was on a {date:%A}'", 'INTERPOLATED_RAW_STRING')
    match("fr'a={d[\"a\"]}'", 'INTERPOLATED_RAW_STRING')
    match("FR'{(lambda x: x*2)(3)}'", 'INTERPOLATED_RAW_STRING')
    match("FR'{(lambda x: x*2)(3)}'", 'INTERPOLATED_RAW_STRING')

    match("rf'He said his name is {name!r}.'", 'INTERPOLATED_RAW_STRING')
    match("rf'The value is {value}.'", 'INTERPOLATED_RAW_STRING')
    match('rF"He said his name is {name!r}."', 'INTERPOLATED_RAW_STRING')
    match('Rf"The value is {value}."', 'INTERPOLATED_RAW_STRING')
    match("RF'{date} was on a {date:%A}'", 'INTERPOLATED_RAW_STRING')
    match("rf'a={d[\"a\"]}'", 'INTERPOLATED_RAW_STRING')
    match("RF'{(lambda x: x*2)(3)}'", 'INTERPOLATED_RAW_STRING')
    match("RF'{(lambda x: x*2)(3)}'", 'INTERPOLATED_RAW_STRING')


def test_raw_string():
    match('r"pouet pouet"', 'RAW_STRING')
    match("r'pouet pouet'", "RAW_STRING")
    match('r"""pouet pouet"""', 'RAW_STRING')
    match("r'''pouet pouet'''", "RAW_STRING")
    match('R"pouet pouet"', 'RAW_STRING')
    match("R'pouet pouet'", "RAW_STRING")
    match('R"""pouet pouet"""', 'RAW_STRING')
    match("R'''pouet pouet'''", "RAW_STRING")


def test_binary_string():
    match('b"pouet pouet"', 'BINARY_STRING')
    match("b'pouet pouet'", "BINARY_STRING")
    match('b"""pouet pouet"""', 'BINARY_STRING')
    match("b'''pouet pouet'''", "BINARY_STRING")
    match('B"pouet pouet"', 'BINARY_STRING')
    match("B'pouet pouet'", "BINARY_STRING")
    match('B"""pouet pouet"""', 'BINARY_STRING')
    match("B'''pouet pouet'''", "BINARY_STRING")


def test_raw_unicode_string():
    match('ur"pouet pouet"', 'UNICODE_RAW_STRING')
    match("uR'pouet pouet'", "UNICODE_RAW_STRING")
    match('uR"""pouet pouet"""', 'UNICODE_RAW_STRING')
    match("ur'''pouet pouet'''", "UNICODE_RAW_STRING")
    match('Ur"pouet pouet"', 'UNICODE_RAW_STRING')
    match("UR'pouet pouet'", "UNICODE_RAW_STRING")
    match('UR"""pouet pouet"""', 'UNICODE_RAW_STRING')
    match("Ur'''pouet pouet'''", "UNICODE_RAW_STRING")


def test_raw_binary_string():
    match('br"pouet pouet"', 'BINARY_RAW_STRING')
    match("bR'pouet pouet'", "BINARY_RAW_STRING")
    match('bR"""pouet pouet"""', 'BINARY_RAW_STRING')
    match("br'''pouet pouet'''", "BINARY_RAW_STRING")
    match('Br"pouet pouet"', 'BINARY_RAW_STRING')
    match("BR'pouet pouet'", "BINARY_RAW_STRING")
    match('BR"""pouet pouet"""', 'BINARY_RAW_STRING')
    match("Br'''pouet pouet'''", "BINARY_RAW_STRING")


def test_big_string_with_endl():
    match('"""pouet\n \r\npouet"""', 'STRING')
    match('u"""pouet\n \r\npouet"""', 'UNICODE_STRING')
    match('r"""pouet\n \r\npouet"""', 'RAW_STRING')
    match('b"""pouet\n \r\npouet"""', 'BINARY_STRING')
    match('ur"""pouet\n \r\npouet"""', 'UNICODE_RAW_STRING')
    match('br"""pouet\n \r\npouet"""', 'BINARY_RAW_STRING')


def test_comment():
    match('# pouet pouet', 'COMMENT')


def test_space():
    match(' ', 'SPACE')
    match('	', 'SPACE')
    match('	 ', 'SPACE')
    match('	 \\\n ', 'SPACE')
    match('	 \\\r\n ', 'SPACE')


def test_arobase():
    match('@', 'AT')


def test_binary():
    match("0b01010101", "BINARY")


def test_zero():
    match("0", "INT")


def test_float_exponant_advanced():
    match("1E1", "FLOAT_EXPONANT")
    match("1E-2", "FLOAT_EXPONANT")
    match("1E+2", "FLOAT_EXPONANT")
    match("1.E1", "FLOAT_EXPONANT")
    match(".1E1", "FLOAT_EXPONANT")
    match("1e1", "FLOAT_EXPONANT")
    match("1e-2", "FLOAT_EXPONANT")
    match("1e+2", "FLOAT_EXPONANT")
    match("1.e1", "FLOAT_EXPONANT")
    match(".1e1", "FLOAT_EXPONANT")


def test_complex():
    match(".1j", "COMPLEX")
    match("1.j", "COMPLEX")
    match("1.1j", "COMPLEX")
    match("11j", "COMPLEX")
    match(".1J", "COMPLEX")
    match("1.J", "COMPLEX")
    match("1.1J", "COMPLEX")
    match("11J", "COMPLEX")


def test_float_advanced():
    match("1.", "FLOAT")
    match(".1", "FLOAT")
    match("1.1", "FLOAT")


def test_exponant_complex():
    match("1e1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.e1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e1j", "FLOAT_EXPONANT_COMPLEX")
    match(".1e1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e1j", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e1j", "FLOAT_EXPONANT_COMPLEX")
    match("1e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e-1j", "FLOAT_EXPONANT_COMPLEX")
    match(".1e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e-1j", "FLOAT_EXPONANT_COMPLEX")
    match("1e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e+1j", "FLOAT_EXPONANT_COMPLEX")
    match(".1e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e+1j", "FLOAT_EXPONANT_COMPLEX")
    match("1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e1J", "FLOAT_EXPONANT_COMPLEX")
    match("1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1e+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1E1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.E1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1E1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1E1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.E1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1E1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1E1J", "FLOAT_EXPONANT_COMPLEX")
    match("1E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1E-1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1E-1J", "FLOAT_EXPONANT_COMPLEX")
    match("1E+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.E+1J", "FLOAT_EXPONANT_COMPLEX")
    match("1.1E+1J", "FLOAT_EXPONANT_COMPLEX")
    match(".1E+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.E+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-1.1E+1J", "FLOAT_EXPONANT_COMPLEX")
    match("-.1E+1J", "FLOAT_EXPONANT_COMPLEX")

# TODO 1.1e1j


def test_backslash():
    match("\\\n", "SPACE")


def test_ellipsis():
    match("...", "ELLIPSIS")


def test_right_arrow():
    match("->", "RIGHT_ARROW")



import errno
import os
import json
import stat
import tempfile
import warnings

from .token import BaronToken

from rply import ParserGenerator
from rply.parser import LRParser
from rply.parsergenerator import LRTable
from rply.errors import ParserGeneratorWarning
from rply.grammar import Grammar
from .utils import BaronError


class ParsingError(BaronError):
    pass


class BaronParserGenerator(ParserGenerator):
    def build(self):
        g = Grammar(self.tokens)

        for level, (assoc, terms) in enumerate(self.precedence, 1):
            for term in terms:
                g.set_precedence(term, assoc, level)

        for prod_name, syms, func, precedence in self.productions:
            g.add_production(prod_name, syms, func, precedence)

        g.set_start()

        for unused_term in g.unused_terminals():
            warnings.warn(
                "Token %r is unused" % unused_term,
                ParserGeneratorWarning,
                stacklevel=2
            )
        for unused_prod in g.unused_productions():
            warnings.warn(
                "Production %r is not reachable" % unused_prod,
                ParserGeneratorWarning,
                stacklevel=2
            )

        g.build_lritems()
        g.compute_first()
        g.compute_follow()

        # win32 temp directories are already per-user
        if os.name == "nt":
            cache_file = os.path.join(
                tempfile.gettempdir(),
                "rply-%s-%s-%s.json" % (self.VERSION, self.cache_id, self.compute_grammar_hash(g))
            )
        else:
            cache_file = os.path.join(
                tempfile.gettempdir(),
                "rply-%s-%s-%s-%s.json" % (self.VERSION, os.getuid(), self.cache_id, self.compute_grammar_hash(g))
            )
        table = None
        if os.path.exists(cache_file):
            with open(cache_file) as f:
                try:
                    data = json.load(f)
                except Exception:
                    os.remove(cache_file)
                    data = None

                if data is not None:
                    stat_result = os.fstat(f.fileno())
                    if (
                        os.name == "nt" or (
                            stat_result.st_uid == os.getuid()
                            and stat.S_IMODE(stat_result.st_mode) == 0o0600
                        )
                    ):
                        if self.data_is_valid(g, data):
                            table = LRTable.from_cache(g, data)

        if table is None:
            table = LRTable.from_grammar(g)
            try:
                fd = os.open(cache_file, os.O_RDWR | os.O_CREAT | os.O_EXCL, 0o0600)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
            else:
                with os.fdopen(fd, "w") as f:
                    json.dump(self.serialize_table(table), f)
        # meh :(
        # if table.sr_conflicts:
            # warnings.warn(
                # "%d shift/reduce conflict%s" % (len(table.sr_conflicts), "s" if len(table.sr_conflicts) > 1 else ""),
                # ParserGeneratorWarning,
                # stacklevel=2,
            # )
        # if table.rr_conflicts:
            # warnings.warn(
                # "%d reduce/reduce conflict%s" % (len(table.rr_conflicts), "s" if len(table.rr_conflicts) > 1 else ""),
                # ParserGeneratorWarning,
                # stacklevel=2,
            # )
        return BaronLRParser(table, self.error_handler)


class BaronLRParser(LRParser):
    def parse(self, tokenizer, state=None):
        lookahead = None
        lookaheadstack = []

        statestack = [0]
        symstack = [BaronToken("$end", "$end")]

        current_state = 0

        parsed_file_content = ""

        while True:
            if self.lr_table.default_reductions[current_state]:
                t = self.lr_table.default_reductions[current_state]
                current_state = self._reduce_production(t, symstack, statestack, state)
                continue

            if lookahead is None:
                if lookaheadstack:
                    lookahead = lookaheadstack.pop()
                else:
                    try:
                        lookahead = next(tokenizer)
                    except StopIteration:
                        lookahead = None

                if lookahead is None:
                    lookahead = BaronToken("$end", "$end")
                else:
                    parsed_file_content += lookahead.render()

            ltype = lookahead.gettokentype()
            if ltype in self.lr_table.lr_action[current_state]:
                t = self.lr_table.lr_action[current_state][ltype]
                if t > 0:
                    statestack.append(t)
                    current_state = t
                    symstack.append(lookahead)
                    lookahead = None
                    continue
                elif t < 0:
                    current_state = self._reduce_production(t, symstack, statestack, state)
                    continue
                else:
                    n = symstack[-1]
                    return n
            else:
                debug_output = parsed_file_content.split("\n")
                debug_output = list(zip(range(1, len(debug_output) + 1), debug_output))
                debug_output = debug_output[-8:]
                debug_output = "\n".join(["%4s %s" % (x[0], x[1]) for x in debug_output])
                debug_output += "<---- here"
                debug_output = "Error, got an unexpected token %s here:\n\n" % ltype + debug_output
                debug_output += "\n\nThe token %s should be one of those: %s" % (ltype, ", ".join(sorted(self.lr_table.lr_action[current_state].keys())))
                debug_output += "\n\nBaron has failed to parse this input. If this is valid python code (and by that I mean that the python binary successfully parse this code without any syntax error) (also consider that baron does not yet parse python 3 code integrally) it would be kind if you can extract a snippet of your code that make Baron fails and open a bug here: https://github.com/PyCQA/baron/issues\n\nSorry for the inconvenience."
                raise ParsingError(debug_output)



import os
import json


def pytest_assertrepr_compare(config, op, left, right):
    if isinstance(left, list) and isinstance(right, list) and op == "==":
        with open("/tmp/a", "w") as a:
            a.write(json.dumps(left, indent=4, sort_keys=True))
        with open("/tmp/b", "w") as b:
            b.write(json.dumps(right, indent=4, sort_keys=True))

        os.system("diff -y /tmp/a /tmp/b")



from .utils import FlexibleIterator, BaronError


class UnExpectedFormattingToken(BaronError):
    pass


class GroupingError(BaronError):
    pass


GROUP_THOSE = (
    "ENDL",

    # TODO test those 2
    "COMMENT",
    "SPACE",
)

ENTER_GROUPING_MODE = (
    "LEFT_PARENTHESIS",
    "LEFT_BRACKET",
    "LEFT_SQUARE_BRACKET",
)

QUIT_GROUPING_MODE = (
    "RIGHT_PARENTHESIS",
    "RIGHT_BRACKET",
    "RIGHT_SQUARE_BRACKET",
)

GROUP_ON = (
    "COMMA",
    "COLON",

    # TODO test everything bellow
    "STRING",
    "RAW_STRING",
    "INTERPOLATED_STRING",
    "INTERPOLATED_RAW_STRING",
    "BINARY_STRING",
    "BINARY_RAW_STRING",
    "UNICODE_STRING",
    "UNICODE_RAW_STRING",

    "AS",
    "IMPORT",
    "DOUBLE_STAR",
    "DOT",
    "LEFT_SQUARE_BRACKET",
    "STAR",
    "SLASH",
    "PERCENT",
    "DOUBLE_SLASH",
    "PLUS",
    "MINUS",
    "LEFT_SHIFT",
    "RIGHT_SHIFT",
    "AMPER",
    "CIRCUMFLEX",
    "VBAR",
    "LESS",
    "GREATER",
    "EQUAL_EQUAL",
    "LESS_EQUAL",
    "GREATER_EQUAL",
    "NOT_EQUAL",
    "IN",
    "IS",
    "NOT",
    "AND",
    "OR",
    "AT",
    "IF",
    "ELSE",
    "FROM",
    "EQUAL",
    "PLUS_EQUAL",
    "MINUS_EQUAL",
    "AT_EQUAL",
    "STAR_EQUAL",
    "SLASH_EQUAL",
    "PERCENT_EQUAL",
    "AMPER_EQUAL",
    "VBAR_EQUAL",
    "CIRCUMFLEX_EQUAL",
    "LEFT_SHIFT_EQUAL",
    "RIGHT_SHIFT_EQUAL",
    "DOUBLE_STAR_EQUAL",
    "DOUBLE_SLASH_EQUAL",
    "ENDL",
    "FOR",
    "COLON",
    "RAW_STRING",
    "UNICODE_STRING",
    "UNICODE_RAW_STRING",
) + ENTER_GROUPING_MODE + QUIT_GROUPING_MODE


def append_to_token_after(token, to_append_list):
    if len(token) == 2:
        return (token[0], token[1], [], to_append_list)
    elif len(token) == 3:
        return (token[0], token[1], token[2], to_append_list)
    elif len(token) == 4:
        return (token[0], token[1], token[2], token[3] + to_append_list)


def append_to_token_before(token, to_append_list):
    if len(token) == 2:
        return (token[0], token[1], to_append_list, [])
    elif len(token) == 3:
        return (token[0], token[1], to_append_list + token[2], [])
    elif len(token) == 4:
        return (token[0], token[1], to_append_list + token[2], token[3])


def group(sequence):
    return list(group_generator(sequence))


def fail_on_bad_token(token, debug_file_content, in_grouping_mode):
    if token[0] in GROUP_ON:
        return

    debug_file_content += _append_to_debug_file_content(token)

    debug_file_content = debug_file_content.split("\n")
    debug_file_content = list(zip(range(1, len(debug_file_content) + 1), debug_file_content))
    debug_file_content = debug_file_content[-8:]
    debug_file_content = "\n".join(["%4s %s" % (x[0], x[1]) for x in debug_file_content])
    raise GroupingError("Fail to group formatting tokens, here:\n%s <----\n\n'%s' should have been in: %s\n\nCurrent value of 'in_grouping_mode': %s" % (debug_file_content, token, ', '.join(sorted(GROUP_ON)), in_grouping_mode))


def _append_to_debug_file_content(token):
    before_debug = "".join(map(lambda x: x[1], token[2] if len(token) >= 3 else []))
    after_debug = "".join(map(lambda x: x[1], token[3] if len(token) >= 4 else []))
    return before_debug + token[1] + after_debug


def group_generator(sequence):
    iterator = FlexibleIterator(sequence)
    current = None, None
    in_grouping_mode = 0
    debug_file_content = ""

    while True:
        if iterator.end():
            return

        current = next(iterator)
        debug_file_content += _append_to_debug_file_content(current)

        if current[0] in ENTER_GROUPING_MODE:
            in_grouping_mode += 1
        elif current[0] in QUIT_GROUPING_MODE:
            in_grouping_mode -= 1
            assert in_grouping_mode >= 0

        if in_grouping_mode:
            if current[0] in GROUP_THOSE:
                to_group = [current]
                while iterator.show_next() and iterator.show_next()[0] in GROUP_THOSE:
                    to_group.append(next(iterator))
                    debug_file_content += _append_to_debug_file_content(to_group[-1])

                    # XXX don't remember how (:() but I can end up finding a
                    # DEDENT/INDENT token in this situation and I don't want to
                    # group on it. Need to do test for that.
                    if iterator.show_next()[0] in ("INDENT", "DEDENT"):
                        yield next(iterator)

                fail_on_bad_token(iterator.show_next(), debug_file_content, in_grouping_mode)

                current = append_to_token_before(next(iterator), to_group)

                if current[0] in ENTER_GROUPING_MODE:
                    in_grouping_mode += 1
                # TODO test
                if current[0] in QUIT_GROUPING_MODE:
                    in_grouping_mode -= 1
                    assert in_grouping_mode >= 0
                    yield current
                    continue

            if current[0] in GROUP_ON:
                while iterator.show_next() and iterator.show_next()[0] in GROUP_THOSE:
                    debug_file_content += _append_to_debug_file_content(iterator.show_next())
                    current = append_to_token_after(current, [next(iterator)])

        yield current



# flake8: noqa

from baron.baron import parse
from baron.path import PathWalker, Position, BoundingBox
from baron.path import position_to_path, path_to_node, position_to_node
from baron.path import path_to_bounding_box, node_to_bounding_box
from baron.utils import string_instance


def test_position():
    pos1 = (1, 2)
    pos2 = Position((1, 2))
    pos3 = Position((1, 2))
    pos4 = Position(Position((1, 2)))

    assert pos1 == pos2
    assert pos2 == pos3
    assert pos3 == pos4

    assert isinstance(pos2, Position)
    assert isinstance(pos3, Position)
    assert isinstance(pos4, Position)


def test_position_compare():
    assert Position((1, 2)) < Position((1, 3))
    assert Position((0, 9)) < Position((1, 3))
    assert Position((1, 2)) <= Position((1, 3))
    assert Position((0, 9)) <= Position((1, 3))
    assert Position((1, 2)) <= Position((1, 2))
    assert Position((1, 2)) == Position((1, 2))

    assert Position((1, 4)) > Position((1, 3))
    assert Position((2, 0)) > Position((1, 3))
    assert Position((1, 4)) >= Position((1, 3))
    assert Position((2, 0)) >= Position((1, 3))
    assert Position((1, 4)) >= Position((1, 4))
    assert Position((1, 4)) == Position((1, 4))


def test_position_bool():
    assert bool(Position((1, 2)))
    assert not bool(Position((-1, 2)))
    assert not bool(Position((1, -2)))
    assert not bool(Position((-1, -2)))


def test_bounding_box():
    bb1 = ((1, 2), (3, 4))
    bb2 = BoundingBox(((1, 2), (3, 4)))
    bb3 = BoundingBox((Position((1, 2)), (3, 4)))
    bb4 = BoundingBox(((1, 2), Position((3, 4))))
    bb5 = BoundingBox((Position((1, 2)), Position((3, 4))))
    bb6 = BoundingBox(((1, 2), (3, 4)))
    bb7 = BoundingBox((Position((1, 2)), (3, 4)))
    bb8 = BoundingBox(((1, 2), Position((3, 4))))
    bb9 = BoundingBox((Position((1, 2)), Position((3, 4))))
    bb10 = BoundingBox(BoundingBox(((1, 2), (3, 4))))

    assert bb1 == bb2
    assert bb2 == bb3
    assert bb3 == bb4
    assert bb4 == bb5
    assert bb5 == bb6
    assert bb6 == bb7
    assert bb7 == bb8
    assert bb8 == bb9
    assert bb9 == bb10

    assert isinstance(bb2, BoundingBox)
    assert isinstance(bb3, BoundingBox)
    assert isinstance(bb4, BoundingBox)
    assert isinstance(bb5, BoundingBox)
    assert isinstance(bb6, BoundingBox)
    assert isinstance(bb7, BoundingBox)
    assert isinstance(bb8, BoundingBox)
    assert isinstance(bb9, BoundingBox)
    assert isinstance(bb10, BoundingBox)


simplecode = """vara = 1"""


bigcode = """\
def fun(arg1, arg2): 
    var *= 1

@deco("arg")
def fun2(arg1 = default, **kwargs):
    arg1.attr = 3
"""


splitcode = """var \\\n  = 2"""
windows_splitcode = """var \\\r\n  = 2"""


classcode = """\
class MyClass(BaseClass):
    def __init__(self, arg1):
        self.a = arg1

    def getA():
        return self.a
"""


class PathWalkerTester(PathWalker):
    """This class is used exclusively to test the PathWalker

    It works by walking along with the PathWalker and for each step,
    i.e. for each before, after and on_leaf method called, it checks if
    that step was indeed expected.
    To specify the expected steps, you should simply pass a list of
    tuples. The first field tells what kind of step it should be:
      * '>' for calling the `before` method, i.e. going down the tree,
      * '<' for calling the `after` method, i.e. going up the tree
    and the second is what the current path should be at that step.
    """
    def __init__(self, paths):
        super(PathWalkerTester, self).__init__(strict=True)
        self.paths = paths

    def before(self, *args):
        super(PathWalkerTester, self).before(*args)
        self.process_test('>')

    def after(self, *args):
        self.process_test('<')
        super(PathWalkerTester, self).after(*args)

    def process_test(self, node_type):
        _node_type, _path = self.paths.pop(0)
        assert _node_type == node_type
        assert _path == self.current_path


def check_path(code, positions, target_path):
    tree = parse(code)

    for position in positions:
        path = position_to_path(tree, position)
        assert path == target_path

        if path is None:
            assert position_to_node(tree, position) is None
            return

        node = path_to_node(tree, path)
        assert isinstance(node, string_instance)

        assert position_to_node(tree, position) is node

    if target_path is not None:
        bounding_box = (positions[0], positions[-1])
        assert path_to_bounding_box(tree, path) == bounding_box


def test_path_walker_assignment():
    node = parse("a = 1")
    # Indentation is purely visual sugar
    walker = PathWalkerTester([
        ('>', [0]),
        ('>', [0, 'target']),
        ('>', [0, 'target', 'value']),
        ('<', [0, 'target', 'value']),
        ('<', [0, 'target']),
        ('>', [0, 'first_formatting']),
        ('>', [0, 'first_formatting', 0]),
        ('>', [0, 'first_formatting', 0, 'value']),
        ('<', [0, 'first_formatting', 0, 'value']),
        ('<', [0, 'first_formatting', 0]),
        ('<', [0, 'first_formatting']),
        ('>', [0, '=']),
        ('<', [0, '=']),
        ('>', [0, 'second_formatting']),
        ('>', [0, 'second_formatting', 0]),
        ('>', [0, 'second_formatting', 0, 'value']),
        ('<', [0, 'second_formatting', 0, 'value']),
        ('<', [0, 'second_formatting', 0]),
        ('<', [0, 'second_formatting']),
        ('>', [0, 'value']),
        ('>', [0, 'value', 'value']),
        ('<', [0, 'value', 'value']),
        ('<', [0, 'value']),
        ('<', [0]),
    ])

    walker.walk(node)


def test_bb_name():
    node = parse("var")[0]
    assert node_to_bounding_box(node) == ((1, 1), (1, 3))


def test_bb_string():
    node = parse("\"hello\"")[0]
    assert path_to_bounding_box(node, []) == ((1, 1), (1, 7))


def test_bb_comment():
    node = parse("# comment")[0]
    assert path_to_bounding_box(node, None) == ((1, 1), (1, 9))


def test_bb_funcdef():
    node = parse(bigcode)[0]
    assert node_to_bounding_box(node) == ((1, 1), (4, 0))


def test_bb_class():
    node = parse(classcode)[0]
    assert node_to_bounding_box(node) == ((1, 1), (7, 0))


def test_bb_split_assignment():
    node = parse(splitcode)[0]
    assert node_to_bounding_box(node) == ((1, 1), (2, 5))


def test_sc_out_of_scope():
    check_path(simplecode, [(-1, 3), (1, -3), (3, 3), (1, 100)], None)


def test_sc_assignement_target():
    path = [0, "target", "value"]
    check_path(simplecode, [(1, 1), (1, 2), (1, 3), (1, 4)], path)


def test_sc_assignement_first_formatting():
    path = [0, "first_formatting", 0, "value"]
    check_path(simplecode, [(1, 5)], path)


def test_sc_assignement_operator():
    path = [0, "="]
    check_path(simplecode, [(1, 6)], path)


def test_sc_assignement_second_formatting():
    path = [0, "second_formatting", 0, "value"]
    check_path(simplecode, [(1, 7)], path)


def test_sc_assignement_value():
    path = [0, "value", "value"]
    check_path(simplecode, [(1, 8)], path)


def test_bc_l1_def():
    path = [0, "def"]
    check_path(bigcode, [(1, 1), (1, 2), (1, 3)], path)


def test_bc_l1_def_first_formatting():
    tree = parse(bigcode)
    path = [0, "first_formatting"]
    assert path_to_bounding_box(tree, path) == ((1, 4), (1, 4))

    path = [0, "first_formatting", 0]
    assert path_to_bounding_box(tree, path) == ((1, 4), (1, 4))

    path = [0, "first_formatting", 0, "value"]
    check_path(bigcode, [(1, 4)], path)


def test_bc_l1_def_name():
    path = [0, "name"]
    check_path(bigcode, [(1, 5), (1, 6), (1, 7)], path)


def test_bc_l1_def_second_formatting():
    path = [0, "second_formatting"]
    tree = parse(bigcode)
    assert path_to_bounding_box(tree, path) == ((1, 8), (1, 7))


def test_bc_l1_def_left_paren():
    path = [0, "("]
    check_path(bigcode, [(1, 8)], path)


def test_bc_l1_arg1():
    path = [0, "arguments", 0, "target", "value"]
    check_path(bigcode, [(1, 9), (1, 10), (1, 11), (1, 12)], path)


def test_bc_l1_comma():
    path = [0, "arguments", 1, ","]
    check_path(bigcode, [(1, 13)], path)


def test_bc_l1_comma_second_formatting():
    path = [0, "arguments", 1, "second_formatting", 0, "value"]
    check_path(bigcode, [(1, 14)], path)


def test_bc_l1_arg2():
    path = [0, "arguments", 2, "target", "value"]
    check_path(bigcode, [(1, 15), (1, 16), (1, 17), (1, 18)], path)


def test_bc_l1_right_paren():
    path = [0, ")"]
    check_path(bigcode, [(1, 19)], path)


def test_bc_l1_colon():
    path = [0, ":"]
    check_path(bigcode, [(1, 20)], path)


def test_bc_l1_space():
    path = [0, "value", 0, "formatting", 0, "value"]
    check_path(bigcode, [(1, 21)], path)


def test_bc_l1_out_of_scope():
    check_path(bigcode, [(1, 22), (2, 0)], None)


def test_bc_l2_endl():
    path = [0, "value", 0, "indent"]
    check_path(bigcode, [(2, 1), (2, 2), (2, 3), (2, 4)], path)


def test_bc_l2_assign_var():
    path = [0, "value", 1, "target", "value"]
    check_path(bigcode, [(2, 5), (2, 6), (2, 7)], path)


def test_bc_l2_assign_first_formatting():
    path = [0, "value", 1, "first_formatting", 0, "value"]
    check_path(bigcode, [(2, 8)], path)


def test_bc_l2_assign_operator():
    path = [0, "value", 1, "operator"]
    check_path(bigcode, [(2, 9)], path)


def test_bc_l2_assign_equal():
    path = [0, "value", 1, "="]
    check_path(bigcode, [(2, 10)], path)


def test_bc_l2_assign_second_formatting():
    path = [0, "value", 1, "second_formatting", 0, "value"]
    check_path(bigcode, [(2, 11)], path)


def test_bc_l2_assign_value():
    path = [0, "value", 1, "value", "value"]
    check_path(bigcode, [(2, 12)], path)


def test_bc_l2_out_of_scope():
    check_path(bigcode, [(2, 13), (3, 0)], None)


def test_bc_l3_empty():
    check_path(bigcode, [(3, 1)], None)


def test_bc_l4_decorator_at():
    path = [1, "decorators", 0, "@"]
    check_path(bigcode, [(4, 1)], path)


def test_bc_l4_decorator_name():
    path = [1, "decorators", 0, "value", "value", 0, "value"]
    check_path(bigcode, [(4, 2), (4, 3), (4, 4), (4, 5)], path)


def test_bc_l4_decorator_left_paren():
    path = [1, "decorators", 0, "call", "("]
    check_path(bigcode, [(4, 6)], path)


def test_bc_l4_decorator_arg1():
    path = [1, "decorators", 0, "call", "value", 0, "value", "value"]
    check_path(bigcode, [(4, 7), (4, 8), (4, 9), (4, 10), (4, 11)], path)


def test_bc_l4_decorator_right_paren():
    path = [1, "decorators", 0, "call", ")"]
    check_path(bigcode, [(4, 12)], path)


def test_bc_l4_out_of_scope():
    check_path(bigcode, [(4, 13), (5, 0)], None)


def test_bc_l5_def():
    path = [1, "def"]
    check_path(bigcode, [(5, 1), (5, 2), (5, 3)], path)


def test_bc_l5_def_first_formatting():
    path = [1, "first_formatting", 0, "value"]
    check_path(bigcode, [(5, 4)], path)


def test_bc_l5_def_name():
    path = [1, "name"]
    check_path(bigcode, [(5, 5), (5, 6), (5, 7), (5, 8)], path)


def test_bc_l5_left_paren():
    path = [1, "("]
    check_path(bigcode, [(5, 9)], path)


def test_bc_l5_arg1():
    path = [1, "arguments", 0, "target", "value"]
    check_path(bigcode, [(5, 10), (5, 11), (5, 12), (5, 13)], path)


def test_bc_l5_arg1_first_formatting():
    path = [1, "arguments", 0, "first_formatting", 0, "value"]
    check_path(bigcode, [(5, 14)], path)


def test_bc_l5_arg1_equal():
    path = [1, "arguments", 0, "="]
    check_path(bigcode, [(5, 15)], path)


def test_bc_l5_arg1_second_formatting():
    path = [1, "arguments", 0, "second_formatting", 0, "value"]
    check_path(bigcode, [(5, 16)], path)


def test_bc_l5_arg1_value():
    path = [1, "arguments", 0, "value", "value"]
    check_path(bigcode, [(5, 17), (5, 18), (5, 19), (5, 20), (5, 21), (5, 22), (5, 23)], path)


def test_bc_l5_args_comma():
    path = [1, "arguments", 1, ","]
    check_path(bigcode, [(5, 24)], path)


def test_bc_l5_args_comma_second_formatting():
    path = [1, "arguments", 1, "second_formatting", 0, "value"]
    check_path(bigcode, [(5, 25)], path)


def test_bc_l5_arg2_operator():
    path = [1, "arguments", 2, "**"]
    check_path(bigcode, [(5, 26), (5, 27)], path)


def test_bc_l5_arg2_name():
    path = [1, "arguments", 2, "value", "value"]
    check_path(bigcode, [(5, 28), (5, 29), (5, 30), (5, 31), (5, 32), (5, 33)], path)


def test_bc_l5_right_paren():
    path = [1, ")"]
    check_path(bigcode, [(5, 34)], path)


def test_bc_l5_colon():
    path = [1, ":"]
    check_path(bigcode, [(5, 35)], path)


def test_bc_l5_out_of_scope():
    check_path(bigcode, [(5, 36), (6, 0)], None)


def test_bc_l6_endl():
    path = [1, "value", 0, "indent"]
    check_path(bigcode, [(6, 1), (6, 2), (6, 3), (6, 4)], path)


def test_bc_l6_assign_var():
    path = [1, "value", 1, "target", "value", 0, "value"]
    check_path(bigcode, [(6, 5), (6, 6), (6, 7), (6, 8)], path)


def test_bc_l6_assign_var_dot():
    path = [1, "value", 1, "target", "value", 1, "."]
    check_path(bigcode, [(6, 9)], path)


def test_bc_l6_assign_var_dot_name():
    path = [1, "value", 1, "target", "value", 2, "value"]
    check_path(bigcode, [(6, 10), (6, 11), (6, 12), (6, 13)], path)


def test_bc_l6_assign_first_formatting():
    path = [1, "value", 1, "first_formatting", 0, "value"]
    check_path(bigcode, [(6, 14)], path)


def test_bc_l6_assign_equal():
    path = [1, "value", 1, "="]
    check_path(bigcode, [(6, 15)], path)


def test_bc_l6_assign_second_formatting():
    path = [1, "value", 1, "second_formatting", 0, "value"]
    check_path(bigcode, [(6, 16)], path)


def test_bc_l6_assign_value():
    path = [1, "value", 1, "value", "value"]
    check_path(bigcode, [(6, 17)], path)


def test_bc_l6_out_of_scope():
    check_path(bigcode, [(6, 18), (7, 0)], None)


def test_sc_l1_var():
    path = [0, "target", "value"]
    check_path(splitcode, [(1, 1), (1, 2), (1, 3)], path)


def test_sc_l1_var_win():
    path = [0, "target", "value"]
    check_path(windows_splitcode, [(1, 1), (1, 2), (1, 3)], path)


def test_sc_l1_l2_first_space():
    path = [0, "first_formatting", 0, "value"]
    check_path(splitcode, [(1, 4), (1, 5), (2, 1), (2, 2)], path)


def test_sc_l1_l2_first_space_win():
    path = [0, "first_formatting", 0, "value"]
    check_path(windows_splitcode, [(1, 4), (1, 5), (2, 1), (2, 2)], path)


def test_sc_l1_out_of_scope():
    check_path(splitcode, [(1, 6), (2, 0)], None)


def test_sc_l1_out_of_scope_win():
    check_path(windows_splitcode, [(1, 6), (2, 0)], None)


def test_sc_l2_operator():
    path = [0, "="]
    check_path(splitcode, [(2, 3)], path)


def test_sc_l2_operator_win():
    path = [0, "="]
    check_path(windows_splitcode, [(2, 3)], path)


def test_sc_l2_second_space():
    path = [0, "second_formatting", 0, "value"]
    check_path(splitcode, [(2, 4)], path)


def test_sc_l2_second_space_win():
    path = [0, "second_formatting", 0, "value"]
    check_path(windows_splitcode, [(2, 4)], path)


def test_sc_l2_value():
    path = [0, "value", "value"]
    check_path(splitcode, [(2, 5)], path)


def test_sc_l2_value_win():
    path = [0, "value", "value"]
    check_path(windows_splitcode, [(2, 5)], path)


def test_sc_l2_out_of_scope():
    check_path(splitcode, [(2, 6), (3, 0)], None)


def test_sc_l2_out_of_scope_win():
    check_path(windows_splitcode, [(2, 6), (3, 0)], None)


def test_position_equality():
    assert Position((1, 2)) == Position((1, 2))
    assert Position((1, 2)) == (1, 2)
    assert Position((1, 2)) != Position((2, 1))
    assert Position((1, 2)) != (2, 1)


def test_position_opposite():
    assert Position((1, 2)) == - Position((-1, -2))


def test_position_arithmetic():
    assert Position((1, 2)) + Position((1, 2)) == (2, 4)
    assert Position((1, 2)) + (1, 2) == (2, 4)
    assert Position((1, 2)) - Position((2, 1)) == (-1, 1)
    assert Position((1, 2)) - (2, 1) == (-1, 1)


def test_position_advance():
    position = Position((1, 1))
    position.advance_columns(3)
    assert position == (1, 4)
    position.advance_line()
    assert position == (2, 1)
    position.advance_columns(4)
    assert position == (2, 5)


def test_position_left():
    assert Position((1, 2)).left == (1, 1)


def test_position_right():
    assert Position((1, 2)).right == (1, 3)


def test_postion_around_the_world():
    assert Position((1, 2)) == Position(Position((1, 2)).to_tuple())



def include_control_structures(pg):
    @pg.production("try_stmt : TRY COLON suite excepts")
    def try_excepts_stmt(pack):
        (try_, colon, suite, excepts) = pack
        return [{
            "type": "try",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
            "else": {},
            "finally": {},
            "excepts": excepts,
        }]

    @pg.production("try_stmt : TRY COLON suite excepts else_stmt")
    def try_excepts_else_stmt(pack):
        (try_, colon, suite, excepts, else_stmt) = pack
        return [{
            "type": "try",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
            "else": else_stmt,
            "finally": {},
            "excepts": excepts,
        }]

    @pg.production("try_stmt : TRY COLON suite excepts finally_stmt")
    def try_excepts_finally_stmt(pack):
        (try_, colon, suite, excepts, finally_stmt) = pack
        return [{
            "type": "try",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
            "else": {},
            "finally": finally_stmt,
            "excepts": excepts,
        }]

    @pg.production("try_stmt : TRY COLON suite excepts else_stmt finally_stmt")
    def try_excepts_else_finally_stmt(pack):
        (try_, colon, suite, excepts, else_stmt, finally_stmt) = pack
        return [{
            "type": "try",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
            "else": else_stmt,
            "finally": finally_stmt,
            "excepts": excepts,
        }]

    @pg.production("try_stmt : TRY COLON suite finally_stmt")
    def try_stmt(pack):
        (try_, colon, suite, finally_stmt) = pack
        return [{
            "type": "try",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
            "else": {},
            "finally": finally_stmt,
            "excepts": [],
        }]

    @pg.production("excepts : excepts except_stmt")
    def excepts(pack):
        (excepts_, except_stmt) = pack
        return excepts_ + except_stmt

    @pg.production("excepts : except_stmt")
    def excepts_except_stmt(pack):
        (except_stmt,) = pack
        return except_stmt

    @pg.production("except_stmt : EXCEPT test AS test COLON suite")
    def except_as_stmt(pack):
        (except_, test, as_, test2, colon, suite) = pack
        return [{
            "type": "except",
            "first_formatting": except_.hidden_tokens_after,
            "second_formatting": as_.hidden_tokens_before,
            "third_formatting": as_.hidden_tokens_after,
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
            "delimiter": "as",
            "target": test2,
            "exception": test,
            "value": suite
        }]

    @pg.production("except_stmt : EXCEPT test COMMA test COLON suite")
    def except_comma_stmt(pack):
        (except_, test, comma, test2, colon, suite) = pack
        return [{
            "type": "except",
            "first_formatting": except_.hidden_tokens_after,
            "second_formatting": comma.hidden_tokens_before,
            "third_formatting": comma.hidden_tokens_after,
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
            "delimiter": ",",
            "target": test2,
            "exception": test,
            "value": suite
        }]

    @pg.production("except_stmt : EXCEPT COLON suite")
    def except_stmt_empty(pack):
        (except_, colon, suite) = pack
        return [{
            "type": "except",
            "first_formatting": except_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
            "delimiter": "",
            "target": {},
            "exception": {},
            "value": suite
        }]

    @pg.production("except_stmt : EXCEPT test COLON suite")
    def except_stmt(pack):
        (except_, test, colon, suite) = pack
        return [{
            "type": "except",
            "first_formatting": except_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
            "delimiter": "",
            "target": {},
            "exception": test,
            "value": suite
        }]

    @pg.production("finally_stmt : FINALLY COLON suite")
    def finally_stmt(pack):
        (finally_, colon, suite) = pack
        return {
            "type": "finally",
            "value": suite,
            "first_formatting": colon.hidden_tokens_before,
            "second_formatting": colon.hidden_tokens_after,
        }

    @pg.production("else_stmt : ELSE COLON suite")
    def else_stmt(pack):
        (else_, colon, suite) = pack
        return {
            "type": "else",
            "value": suite,
            "first_formatting": else_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_after,
        }

    @pg.production("for_stmt : FOR exprlist IN testlist COLON suite")
    def for_stmt(pack,):
        (for_, exprlist, in_, testlist, colon, suite) = pack
        return [{
            "type": "for",
            "async": False,
            "async_formatting": [] + for_.hidden_tokens_before,
            "value": suite,
            "iterator": exprlist,
            "target": testlist,
            "else": {},
            "first_formatting": for_.hidden_tokens_after,
            "second_formatting": in_.hidden_tokens_before,
            "third_formatting": in_.hidden_tokens_after,
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
        }]

    @pg.production("for_stmt : FOR exprlist IN testlist COLON suite else_stmt")
    def for_else_stmt(pack,):
        (for_, exprlist, in_, testlist, colon, suite, else_stmt) = pack
        return [{
            "type": "for",
            "value": suite,
            "async": False,
            "async_formatting": [] + for_.hidden_tokens_before,
            "iterator": exprlist,
            "target": testlist,
            "else": else_stmt,
            "first_formatting": for_.hidden_tokens_after,
            "second_formatting": in_.hidden_tokens_before,
            "third_formatting": in_.hidden_tokens_after,
            "fourth_formatting": colon.hidden_tokens_before,
            "fifth_formatting": colon.hidden_tokens_after,
        }]

    @pg.production("while_stmt : WHILE test COLON suite")
    def while_stmt(pack):
        (while_, test, colon, suite) = pack
        return [{
            "type": "while",
            "value": suite,
            "test": test,
            "else": {},
            "first_formatting": while_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
        }]

    @pg.production("while_stmt : WHILE test COLON suite else_stmt")
    def while_stmt_else(pack):
        (while_, test, colon, suite, else_stmt) = pack
        return [{
            "type": "while",
            "value": suite,
            "test": test,
            "else": else_stmt,
            "first_formatting": while_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
        }]

    @pg.production("if_stmt : IF test COLON suite")
    def if_stmt(pack):
        (if_, test, colon, suite) = pack
        return [{
            "type": "ifelseblock",
            "value": [{
                "type": "if",
                "value": suite,
                "test": test,
                "first_formatting": if_.hidden_tokens_after,
                "second_formatting": colon.hidden_tokens_before,
                "third_formatting": colon.hidden_tokens_after,
            }]
        }]

    @pg.production("if_stmt : IF test COLON suite elifs")
    def if_elif_stmt(pack):
        (if_, test, colon, suite, elifs) = pack
        return [{
            "type": "ifelseblock",
            "value": [{
                "type": "if",
                "value": suite,
                "test": test,
                "first_formatting": if_.hidden_tokens_after,
                "second_formatting": colon.hidden_tokens_before,
                "third_formatting": colon.hidden_tokens_after,
            }] + elifs
        }]

    @pg.production("elifs : elifs ELIF test COLON suite")
    def elifs_elif(pack,):
        (elifs, elif_, test, colon, suite) = pack
        return elifs + [{
            "type": "elif",
            "first_formatting": elif_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
            "value": suite,
            "test": test,
        }]

    @pg.production("elifs : ELIF test COLON suite")
    def elif_(pack,):
        (elif_, test, colon, suite) = pack
        return [{
            "type": "elif",
            "first_formatting": elif_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
            "value": suite,
            "test": test,
        }]

    @pg.production("if_stmt : IF test COLON suite else_stmt")
    def if_else_stmt(pack):
        (if_, test, colon, suite, else_stmt) = pack
        return [{
            "type": "ifelseblock",
            "value": [{
                "type": "if",
                "value": suite,
                "test": test,
                "first_formatting": if_.hidden_tokens_after,
                "second_formatting": colon.hidden_tokens_before,
                "third_formatting": colon.hidden_tokens_after,
            }, else_stmt]
        }]

    @pg.production("if_stmt : IF test COLON suite elifs else_stmt")
    def if_elif_else_stmt(pack):
        (if_, test, colon, suite, elifs, else_stmt) = pack
        return [{
            "type": "ifelseblock",
            "value": [{
                "type": "if",
                "value": suite,
                "test": test,
                "first_formatting": if_.hidden_tokens_after,
                "second_formatting": colon.hidden_tokens_before,
                "third_formatting": colon.hidden_tokens_after,
            }] + elifs + [else_stmt]
        }]



import sys

from baron.grammator import generate_parse
from baron.dumper import dumps
from baron.baron import parse as baron_parse
from baron.utils import python_version

if python_version == 3:
    from itertools import zip_longest
else:
    from itertools import izip_longest
    zip_longest = izip_longest

parse = generate_parse(False)


def parse_simple(tokens, result):
    if not tokens or tokens[-1][0] != "ENDL":
        tokens += [('ENDL', '\n')]
    assert parse(tokens + [('ENDMARKER', ''), None]) == (result + [{"type": "endl", "value": "\n", "formatting": [], "indent": ""}])


def parse_multi(tokens, result):
    assert parse(tokens + [('ENDMARKER', ''), None]) == result


def check_dumps(source_code):
    try:
        open("/tmp/c", "w").write(source_code)
        open("/tmp/d", "w").write(dumps(baron_parse(source_code)))
    except Exception as e:
        import json
        import traceback
        traceback.print_exc(file=sys.stdout)
        sys.stdout.write("Warning: couldn't write dumps output to debug file, exception: %s\n\n" % e)
        sys.stdout.write("Tree: %s" % json.dumps(baron_parse(source_code), indent=4) + "\n")

    assert dumps(baron_parse(source_code), strict=True) == source_code



# encoding: utf-8

from .test_utils import parse_simple


def test_empty_tuple():
    "()"
    parse_simple([
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
    ], [
        {
            "with_parenthesis": True,
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "type": "tuple",
            "value": [],
        }
    ])


def test_empty_tuple_space():
    "(  )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', '  ')]),
        ('RIGHT_PARENTHESIS', ')'),
    ], [
        {
            "with_parenthesis": True,
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": "  "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "type": "tuple",
            "value": [],
        }
    ])


def test_associative_parenthesis():
    "( a )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "associative_parenthesis",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "value": {
                "type": "name",
                "value": "a",
            },
        }
    ])


def test_tuple_one():
    "( a, )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')', [], []),
    ], [
        {
            "with_parenthesis": True,
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "type": "tuple",
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                }
            ],
        }
    ])


def test_tuple_many():
    "(a, b, c)"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], []),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')', [], []),
    ], [
        {
            "with_parenthesis": True,
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "type": "tuple",
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "b",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "c",
                }
            ],
        }
    ])


def test_empty_list():
    "[ ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('RIGHT_SQUARE_BRACKET', ']', [], []),
    ], [
        {
            "type": "list",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [],
        }
    ])


def test_list_one():
    "[ a ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                }
            ],
        }
    ])


def test_list_more():
    "[a, b, c]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], []),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_SQUARE_BRACKET', ']', [], []),
    ], [
        {
            "type": "list",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "b",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "c",
                }
            ],
        }
    ])


def test_dict_empty():
    "{ }"
    parse_simple([
        ('LEFT_BRACKET', '{', [], [('SPACE', ' ')]),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "dict",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [],
        }
    ])


def test_dict_one_colon():
    "{a: b}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "dict",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "dictitem",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "key": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": {
                        "type": "name",
                        "value": "b",
                    }
                }
            ],
        }
    ])


def test_dict_one_double_star():
    "{**a}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('DOUBLE_STAR', '**'),
        ('NAME', 'a'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            'type': 'dict',
            'first_formatting': [],
            'second_formatting': [],
            'third_formatting': [],
            'fourth_formatting': [],
            'value': [
                {
                    'type': 'dict_argument',
                    'annotation': {},
                    'annotation_first_formatting': [],
                    'annotation_second_formatting': [],
                    'formatting': [],
                    'value': {'type': 'name', 'value': 'a'}
                }
            ]
        }
    ]
)


def test_dict_more_colon():
    "{a: b, b: c, c: d}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "dict",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "dictitem",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "key": {
                        "type": "name",
                        "value": "a",
                    },
                    "value": {
                        "type": "name",
                        "value": "b",
                    }
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "dictitem",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "key": {
                        "type": "name",
                        "value": "b",
                    },
                    "value": {
                        "type": "name",
                        "value": "c",
                    }
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "dictitem",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "key": {
                        "type": "name",
                        "value": "c",
                    },
                    "value": {
                        "type": "name",
                        "value": "d",
                    }
                }
            ],
        }
    ])


def test_dict_more_double_star():
    "{**a, b: c}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('DOUBLE_STAR', '**'),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "first_formatting": [],
            "fourth_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "type": "dict",
            "value": [
                {
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "formatting": [],
                    "type": "dict_argument",
                    "value": {
                        "type": "name",
                        "value": "a"
                    }
                },
                {
                    "first_formatting": [],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "comma"
                },
                {
                    "first_formatting": [],
                    "key": {
                        "type": "name",
                        "value": "b"
                    },
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "dictitem",
                    "value": {
                        "type": "name",
                        "value": "c"
                    }
                }
            ]
        }
    ])


def test_set_one():
    "{a}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('NAME', 'a'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "set",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                }
            ],
        }
    ])


def test_set_one_comma():
    "{a,}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('NAME', 'a'),
        ('COMMA', ','),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "set",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [],
                }
            ],
        }
    ])


def test_set_more():
    "{a, b, c}"
    parse_simple([
        ('LEFT_BRACKET', '{', [], []),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_BRACKET', '}', [], []),
    ], [
        {
            "type": "set",
            "first_formatting": [],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "b",
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "name",
                    "value": "c",
                }
            ],
        }
    ])


def test_generator_comprehension():
    "( a for b in c )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "generator_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_generator_comprehension_if():
    "( a for b in c if d )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "generator_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_generator_comprehension_if_if():
    "( a for b in c if d if e )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'e'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "generator_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        },
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "e"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_generator_comprehension_double():
    "( a for b in c for d in e )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'e'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "generator_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [],
                },
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "d",
                    },
                    "target": {
                        "type": "name",
                        "value": "e",
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_generator_comprehension_double_if_if():
    "( a for b in c if x for d in e if y )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'e'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'y'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "generator_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "x"
                            },
                        }
                    ],
                },
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "d",
                    },
                    "target": {
                        "type": "name",
                        "value": "e",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "y"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_list_comprehension():
    "[ a for b in c ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_list_comprehension_if():
    "[ a for b in c if d ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_list_comprehension_if_if():
    "[ a for b in c if d if d ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        },
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_list_comprehension_tuple():
    "[ a for b in c, d ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "tuple",
                        "with_parenthesis": False,
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "value": [
                            {
                                "type": "name",
                                "value": "c",
                            },
                            {
                                "type": "comma",
                                "first_formatting": [],
                                "second_formatting": [{"type": "space", "value": " "}],
                            },
                            {
                                "type": "name",
                                "value": "d",
                            }
                        ],
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_list_comprehension_tuple_more():
    "[ a for b in c, d, e ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'e'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "tuple",
                        "with_parenthesis": False,
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "value": [
                            {
                                "type": "name",
                                "value": "c",
                            },
                            {
                                "type": "comma",
                                "first_formatting": [],
                                "second_formatting": [{"type": "space", "value": " "}],
                            },
                            {
                                "type": "name",
                                "value": "d",
                            },
                            {
                                "type": "comma",
                                "first_formatting": [],
                                "second_formatting": [{"type": "space", "value": " "}],
                            },
                            {
                                "type": "name",
                                "value": "e",
                            }
                        ],
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_list_comprehension_lambda():
    "[ a for b in lambda: c ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('LAMBDA', 'lambda', [], []),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "lambda",
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [{"type": "space", "value": " "}],
                        "arguments": [],
                        "value": {
                            "type": "name",
                            "value": "c",
                        }
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_list_comprehension_lambda_with_arguments():
    "[ a for b in lambda argument: c ]"
    parse_simple([
        ('LEFT_SQUARE_BRACKET', '[', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('LAMBDA', 'lambda', [], [('SPACE', ' ')]),
        ('NAME', 'argument'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_SQUARE_BRACKET', ']', [('SPACE', ' ')]),
    ], [
        {
            "type": "list_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "lambda",
                        "first_formatting": [{"type": "space", "value": " "}],
                        "second_formatting": [],
                        "third_formatting": [{"type": "space", "value": " "}],
                        "arguments": [
                            {
                                "value": {},
                                "first_formatting": [],
                                "second_formatting": [],
                                "annotation": {},
                                "annotation_first_formatting": [],
                                "annotation_second_formatting": [],
                                "type": "def_argument",
                                "target": {
                                    "type": "name",
                                    "value": "argument",
                                }
                            }
                        ],
                        "value": {
                            "type": "name",
                            "value": "c",
                        }
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_set_comprehension():
    "{ a for b in c }"
    parse_simple([
        ('LEFT_BRACKET', '{', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_BRACKET', '}', [('SPACE', ' ')]),
    ], [
        {
            "type": "set_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_set_comprehension_if():
    "{ a for b in c if d }"
    parse_simple([
        ('LEFT_BRACKET', '{', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('RIGHT_BRACKET', '}', [('SPACE', ' ')]),
    ], [
        {
            "type": "set_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "d"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_set_comprehension_double_if_if():
    "{ a for b in c if x for d in e if y }"
    parse_simple([
        ('LEFT_BRACKET', '{', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'd'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'e'),
        ('IF', 'if', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'y'),
        ('RIGHT_BRACKET', '}', [('SPACE', ' ')]),
    ], [
        {
            "type": "set_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "type": "name",
                "value": "a",
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "x"
                            },
                        }
                    ],
                },
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "d",
                    },
                    "target": {
                        "type": "name",
                        "value": "e",
                    },
                    "ifs": [
                        {
                            "type": "comprehension_if",
                            "first_formatting": [{"type": "space", "value": " "}],
                            "second_formatting": [{"type": "space", "value": " "}],
                            "value": {
                                "type": "name",
                                "value": "y"
                            },
                        }
                    ],
                }
            ]
        }
    ])


def test_dict_comprehension():
    "{ a: x for b in c }"
    parse_simple([
        ('LEFT_BRACKET', '{', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('FOR', 'for', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('IN', 'in', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('RIGHT_BRACKET', '}', [('SPACE', ' ')]),
    ], [
        {
            "type": "dict_comprehension",
            "first_formatting": [],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "result": {
                "key": {
                    "type": "name",
                    "value": "a",
                },
                "first_formatting": [],
                "type": "dictitem",
                "second_formatting": [{"type": "space", "value": " "}],
                "value": {
                    "type": "name",
                    "value": "x",
                },
            },
            "generators": [
                {
                    "type": "comprehension_loop",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "third_formatting": [{"type": "space", "value": " "}],
                    "fourth_formatting": [{"type": "space", "value": " "}],
                    "iterator": {
                        "type": "name",
                        "value": "b",
                    },
                    "target": {
                        "type": "name",
                        "value": "c",
                    },
                    "ifs": [],
                }
            ]
        }
    ])


def test_yield_atom_empty():
    "( yield )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('YIELD', 'yield', [], [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')', []),
    ], [
        {
            "type": "yield_atom",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "value": None,
        }
    ])


def test_yield_atom():
    "( yield a )"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('SPACE', ' ')]),
        ('YIELD', 'yield', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
    ], [
        {
            "type": "yield_atom",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "value": {
                "type": "name",
                "value": "a",
            },
        }
    ])


def test_repr_quote():
    "` a `"
    parse_simple([
        ('BACKQUOTE', '`', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('BACKQUOTE', '`', [('SPACE', ' ')]),
    ], [
        {
            "type": "repr",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                }
            ],
        }
    ])


def test_repr_quote_double():
    "` a, b `"
    parse_simple([
        ('BACKQUOTE', '`', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('BACKQUOTE', '`', [('SPACE', ' ')]),
    ], [
        {
            "type": "repr",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "value": [
                {
                    "type": "name",
                    "value": "a",
                },
                {
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "type": "comma",
                },
                {
                    "type": "name",
                    "value": "b",
                }
            ],
        }
    ])


def test_empty_tuple_endl():
    "(\n)"
    parse_simple([
        ('LEFT_PARENTHESIS', '(', [], [('ENDL', '\n')]),
        ('RIGHT_PARENTHESIS', ')'),
    ], [
        {
            "with_parenthesis": True,
            "first_formatting": [],
            "second_formatting": [
                {
                    "indent": "",
                    "formatting": [],
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "third_formatting": [],
            "fourth_formatting": [],
            "type": "tuple",
            "value": [],
        }
    ])


def test_subscript_special_case():
    "z[a, b, c,]"
    parse_simple([
        ('NAME', 'z'),
        ('LEFT_SQUARE_BRACKET', '['),
        ('NAME', 'a'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COMMA', ','),
        ('RIGHT_SQUARE_BRACKET', ']')
    ], [
        {
            "type": "atomtrailers",
            "value": [
                {
                    "type": "name",
                    "value": "z"
                },
                {
                    "first_formatting": [],
                    "type": "getitem",
                    "value": {
                        "first_formatting": [],
                        "with_parenthesis": False,
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "type": "tuple",
                        "value": [
                            {
                                "type": "name",
                                "value": "a"
                            },
                            {
                                "first_formatting": [],
                                "type": "comma",
                                "second_formatting": [
                                    {
                                        "type": "space",
                                        "value": " "
                                    }
                                ]
                            },
                            {
                                "type": "name",
                                "value": "b"
                            },
                            {
                                "first_formatting": [],
                                "type": "comma",
                                "second_formatting": [
                                    {
                                        "type": "space",
                                        "value": " "
                                    }
                                ]
                            },
                            {
                                "type": "name",
                                "value": "c"
                            },
                            {
                                "first_formatting": [],
                                "type": "comma",
                                "second_formatting": []
                            }
                        ],
                        "second_formatting": []
                    },
                    "second_formatting": [],
                    "third_formatting": [],
                    "fourth_formatting": []
                }
            ]
        }
    ])



from .utils import FlexibleIterator, BaronError


class UnExpectedSpaceToken(BaronError):
    pass


PRIORITY_ORDER = (
    "IMPORT",
    "ENDL",
)

BOTH = (
    "SEMICOLON",
    "AS",
    "IMPORT",
    "DOUBLE_STAR",
    "DOT",
    "LEFT_SQUARE_BRACKET",
    "LEFT_PARENTHESIS",
    "STAR",
    "SLASH",
    "PERCENT",
    "DOUBLE_SLASH",
    "PLUS",
    "MINUS",
    "AT",
    "LEFT_SHIFT",
    "RIGHT_SHIFT",
    "AMPER",
    "CIRCUMFLEX",
    "VBAR",
    "LESS",
    "GREATER",
    "EQUAL_EQUAL",
    "LESS_EQUAL",
    "GREATER_EQUAL",
    "NOT_EQUAL",
    "IN",
    "IS",
    "NOT",
    "AND",
    "OR",
    "IF",
    "ELSE",
    "EQUAL",
    "PLUS_EQUAL",
    "MINUS_EQUAL",
    "STAR_EQUAL",
    "AT_EQUAL",
    "SLASH_EQUAL",
    "PERCENT_EQUAL",
    "AMPER_EQUAL",
    "VBAR_EQUAL",
    "CIRCUMFLEX_EQUAL",
    "LEFT_SHIFT_EQUAL",
    "RIGHT_SHIFT_EQUAL",
    "DOUBLE_STAR_EQUAL",
    "DOUBLE_SLASH_EQUAL",
    "ENDL",
    "COMMA",
    "FOR",
    "COLON",
    "BACKQUOTE",
    "RIGHT_ARROW",
    "FROM",
)

STRING = (
    "STRING",
    "RAW_STRING",
    "INTERPOLATED_STRING",
    "INTERPOLATED_RAW_STRING",
    "UNICODE_STRING",
    "UNICODE_RAW_STRING",
    "BINARY_STRING",
    "BINARY_RAW_STRING",
)

GROUP_SPACE_BEFORE = BOTH + (
    "RIGHT_PARENTHESIS",
    "COMMENT",
) + STRING

GROUP_SPACE_AFTER = BOTH + (
    "TILDE",
    "RETURN",
    "YIELD",
    "WITH",
    "DEL",
    "ASSERT",
    "RAISE",
    "EXEC",
    "GLOBAL",
    "NONLOCAL",
    "PRINT",
    "INDENT",
    "WHILE",
    "ELIF",
    "EXCEPT",
    "DEF",
    "CLASS",
    "LAMBDA",
)


def less_prioritary_than(a, b):
    if b not in PRIORITY_ORDER:
        return False

    if a not in PRIORITY_ORDER:
        return True

    return PRIORITY_ORDER.index(a) < PRIORITY_ORDER.index(b)


def group(sequence):
    return list(group_generator(sequence))


def group_generator(sequence):
    iterator = FlexibleIterator(sequence)

    while not iterator.end():
        current = next(iterator)

        if current is None:
            return

        if current[0] == "SPACE" and iterator.show_next() and iterator.show_next()[0] in GROUP_SPACE_BEFORE:
            new_current = next(iterator)
            current = (new_current[0], new_current[1], [current])

        if current[0] in GROUP_SPACE_AFTER + STRING and\
            (iterator.show_next() and iterator.show_next()[0] == "SPACE") and\
                (not iterator.show_next(2) or (iterator.show_next(2) and not less_prioritary_than(current[0], iterator.show_next(2)[0]))):

            # do not be greedy when you are grouping on strings
            if current[0] in STRING and iterator.show_next(2) and iterator.show_next(2)[0] in GROUP_SPACE_BEFORE:
                yield current
                continue

            after_space = next(iterator)
            current = (current[0], current[1], current[2] if len(current) > 2 else [], [after_space])

        # in case of "def a():  # comment\n pass"
        # not really happy about this solution but that avoid a broken release
        if current[0] == "COLON" and iterator.show_next() and iterator.show_next()[0] == "COMMENT":
            comment = next(iterator)
            current = (current[0], current[1], ((current[2]) if len(current) > 2 else []), ((current[3]) if len(current) > 3 else []) + [comment])

        yield current



from .utils import create_node_from_token


def include_imports(pg):
    @pg.production("small_stmt : import")
    @pg.production("small_stmt : from_import")
    def separator(pack):
        (statement,) = pack
        return statement

    @pg.production("import : IMPORT dotted_as_names")
    def importeu(pack):
        (import_, dotted_as_names) = pack
        return {
            "type": "import",
            "value": dotted_as_names,
            "first_formatting": import_.hidden_tokens_before,
            "second_formatting": import_.hidden_tokens_after
        }

    @pg.production("from_import : FROM dotted_name IMPORT from_import_target")
    def from_import_with_space(pack):
        (from_, dotted_name, import_, from_import_target) = pack
        return {
            "type": "from_import",
            "targets": from_import_target,
            "first_formatting": from_.hidden_tokens_after,
            "second_formatting": import_.hidden_tokens_before,
            "third_formatting": import_.hidden_tokens_after,
            "value": dotted_name
        }

    @pg.production("from_import_target : name_as_names")
    def from_import_target_name_as_names(pack):
        (name_as_names,) = pack
        return name_as_names

    @pg.production("from_import_target : LEFT_PARENTHESIS name_as_names RIGHT_PARENTHESIS")
    def from_import_parenthesis(pack):
        (left_parenthesis, name_as_names, right_parenthesis) = pack
        return left_parenthesis.hidden_tokens_before +\
            [{"type": "left_parenthesis", "value": "("}] +\
            left_parenthesis.hidden_tokens_after +\
            name_as_names +\
            right_parenthesis.hidden_tokens_before +\
            [{"type": "right_parenthesis", "value": ")"}] +\
            right_parenthesis.hidden_tokens_after

    @pg.production("from_import_target : STAR")
    def from_import_star(pack):
        (star,) = pack
        return [{
            "type": "star",
            "value": "*",
            "first_formatting": star.hidden_tokens_before,
            "second_formatting": star.hidden_tokens_after
        }]

    @pg.production("name_as_names : name_as_names name_as_name")
    def name_as_names_name_as_name(pack):
        (name_as_names, name_as_name) = pack
        return name_as_names + name_as_name

    @pg.production("name_as_names : name_as_name")
    def name_as_names(pack):
        (name_as_name,) = pack
        return name_as_name

    @pg.production("name_as_name : NAME AS NAME")
    def name_as_name_name_as_name(pack):
        (name, as_, name2) = pack
        return [{
            "type": "name_as_name",
            "value": name.value,
            "first_formatting": as_.hidden_tokens_before,
            "second_formatting": as_.hidden_tokens_after,
            "target": name2.value
        }]

    @pg.production("name_as_name : NAME")
    def name_as_name_name(pack):
        (name,) = pack
        return [{
            "type": "name_as_name",
            "value": name.value,
            "target": "",
            "first_formatting": [],
            "second_formatting": []
        }]

    @pg.production("name_as_name : NAME SPACE")
    def name_as_name_name_space(pack):
        (name, space) = pack
        return [{
            "type": "name_as_name",
            "target": None,
            "value": name.value,
            "first_formatting": [],
            "second_formatting": []
        }] + [create_node_from_token(space)]

    @pg.production("name_as_name : comma")
    def name_as_name_comma_space(pack):
        (comma,) = pack
        return [comma]

    @pg.production("dotted_as_names : dotted_as_names comma dotted_as_name")
    def dotted_as_names_dotted_as_names_dotted_as_name(pack):
        (dotted_as_names, comma, dotted_as_names2) = pack
        return dotted_as_names + [comma] + dotted_as_names2

    @pg.production("dotted_as_names : dotted_as_name")
    def dotted_as_names_dotted_as_name(pack):
        (dotted_as_name,) = pack
        return dotted_as_name

    @pg.production("dotted_as_name : dotted_name AS NAME")
    def dotted_as_name_as(pack):
        (dotted_name, as_, name) = pack
        return [{
            "type": "dotted_as_name",
            "value": dotted_name,
            "first_formatting": as_.hidden_tokens_before,
            "second_formatting": as_.hidden_tokens_after,
            "target": name.value,
        }]

    @pg.production("dotted_as_name : dotted_name")
    def dotted_as_name(pack):
        (dotted_name,) = pack
        return [{
            "type": "dotted_as_name",
            "value": dotted_name,
            "first_formatting": [],
            "second_formatting": [],
            "target": ""
        }]

    @pg.production("dotted_name : dotted_name dotted_name_element")
    def dotted_name_elements_element(pack):
        (dotted_name, dotted_name_element) = pack
        return dotted_name + dotted_name_element

    @pg.production("dotted_name : dotted_name_element")
    def dotted_name_element(pack):
        (dotted_name_element,) = pack
        return dotted_name_element

    @pg.production("dotted_name_element : NAME")
    @pg.production("dotted_name_element : SPACE")
    def dotted_name(pack):
        (token,) = pack
        return [create_node_from_token(token)]

    @pg.production("dotted_name_element : DOT")
    def dotted_name_dot(pack):
        (dot,) = pack
        return [{
            "type": "dot",
            "first_formatting": dot.hidden_tokens_before,
            "second_formatting": dot.hidden_tokens_after,
        }]

    @pg.production("dotted_name_element : ELLIPSIS")
    def dotted_name_dot_dot_dot(pack):
        ellipsis = pack[0]
        return [{
            "type": "ellipsis",
            "first_formatting": ellipsis.hidden_tokens_before,
            "second_formatting": ellipsis.hidden_tokens_after,
        }]
